/*
 * Copyright (C) 2012  Dan Nicholson <dbn.lists@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.

 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA.
 */

#include <config.h>
#include "evbp-viewer.h"
#include "ev-page-action.h"
#include <gtk/gtk.h>

G_DEFINE_TYPE(EvbpViewer, evbp_viewer, GTK_TYPE_BOX)

static void
evbp_viewer_zoom_best_fit(GtkToggleAction *action, EvbpViewer *viewer)
{
    ev_document_model_set_sizing_mode(viewer->model,
                                      gtk_toggle_action_get_active(action) ?
                                      EV_SIZING_BEST_FIT : EV_SIZING_FREE);
}

static void
evbp_viewer_zoom_page_width(GtkToggleAction *action, EvbpViewer *viewer)
{
    ev_document_model_set_sizing_mode(viewer->model,
                                      gtk_toggle_action_get_active(action) ?
                                      EV_SIZING_FIT_WIDTH : EV_SIZING_FREE);
}

static void
evbp_viewer_toggle_continuous(GtkToggleAction *action, EvbpViewer *viewer)
{
    ev_document_model_set_continuous(viewer->model,
                                     gtk_toggle_action_get_active(action) ?
                                     TRUE : FALSE);
}

static void
evbp_viewer_previous_page(GtkAction *action, EvbpViewer *viewer)
{
    ev_view_previous_page(EV_VIEW(viewer->view));
}

static void
evbp_viewer_next_page(GtkAction *action, EvbpViewer *viewer)
{
    ev_view_next_page(EV_VIEW(viewer->view));
}

static void
sizing_mode_handler(GObject *gobject, GParamSpec *pspec, gpointer user_data)
{
    EvDocumentModel *model = EV_DOCUMENT_MODEL(gobject);
    EvbpViewer *viewer = EVBP_VIEWER(user_data);
    EvSizingMode mode;
    GtkAction *action;

    /* set toggle states based on model */
    mode = ev_document_model_get_sizing_mode(model);

    action = gtk_action_group_get_action(viewer->action_group,
                                         "ViewBestFit");
    g_signal_handlers_block_by_func(action,
                                    G_CALLBACK(evbp_viewer_zoom_best_fit),
                                    viewer);
    gtk_toggle_action_set_active(GTK_TOGGLE_ACTION(action),
                                 mode == EV_SIZING_BEST_FIT);
    g_signal_handlers_unblock_by_func(action,
                                      G_CALLBACK(evbp_viewer_zoom_best_fit),
                                      viewer);

    action = gtk_action_group_get_action(viewer->action_group,
                                         "ViewPageWidth");
    g_signal_handlers_block_by_func(action,
                                    G_CALLBACK(evbp_viewer_zoom_page_width),
                                    viewer);
    gtk_toggle_action_set_active(GTK_TOGGLE_ACTION(action),
                                 mode == EV_SIZING_FIT_WIDTH);
    g_signal_handlers_unblock_by_func(action,
                                      G_CALLBACK(evbp_viewer_zoom_page_width),
                                      viewer);
}

static void
continuous_handler(GObject *gobject, GParamSpec *pspec, gpointer user_data)
{
    EvDocumentModel *model = EV_DOCUMENT_MODEL(gobject);
    EvbpViewer *viewer = EVBP_VIEWER(user_data);
    GtkAction *action;

    /* set toggle state based on model */
    action = gtk_action_group_get_action(viewer->action_group,
                                         "ViewContinuous");
    g_signal_handlers_block_by_func(action,
                                    G_CALLBACK(evbp_viewer_toggle_continuous),
                                    viewer);
    gtk_toggle_action_set_active(GTK_TOGGLE_ACTION(action),
                                 ev_document_model_get_continuous(model));
    g_signal_handlers_unblock_by_func(action,
                                      G_CALLBACK(evbp_viewer_toggle_continuous),
                                      viewer);
}

static void
page_action_link_activated(GtkAction *action, GObject *obj, gpointer user_data)
{
    EvLink *link = EV_LINK(obj);
    EvbpViewer *viewer = EVBP_VIEWER(user_data);

    ev_view_handle_link(EV_VIEW(viewer->view), link);
    gtk_widget_grab_focus(viewer->view);
}

static void
focus_page_selector(GtkAction *action, EvbpViewer *viewer)
{
    GtkAction *page_action;

    page_action = gtk_action_group_get_action(viewer->action_group,
                                              "PageSelector");
    ev_page_action_grab_focus(EV_PAGE_ACTION(page_action));
}

static gboolean
view_focus_changed(GtkWidget *widget, GdkEvent *event, gpointer user_data)
{
    GdkEventFocus *focus = (GdkEventFocus *)event;
    EvbpViewer *viewer = EVBP_VIEWER(user_data);

    if (viewer->accel_group)
        gtk_action_group_set_sensitive(viewer->accel_group, focus->in);

    return FALSE;
}

static void
evbp_viewer_dispose(GObject *object)
{
    EvbpViewer *viewer = EVBP_VIEWER(object);

    if (viewer->document) {
        g_object_unref(viewer->document);
        viewer->document = NULL;
    }
    if (viewer->model) {
        g_object_unref(viewer->model);
        viewer->model = NULL;
    }
    if (viewer->action_group) {
        g_object_unref(viewer->action_group);
        viewer->action_group = NULL;
    }
    if (viewer->accel_group) {
        g_object_unref(viewer->accel_group);
        viewer->accel_group = NULL;
    }
    if (viewer->ui_manager) {
        g_object_unref(viewer->ui_manager);
        viewer->ui_manager = NULL;
    }

    G_OBJECT_CLASS(evbp_viewer_parent_class)->dispose(object);
}

static void
evbp_viewer_hierarchy_changed(GtkWidget *widget, GtkWidget *previous_toplevel)
{
    EvbpViewer *viewer = EVBP_VIEWER(widget);
    GtkWidget *toplevel;
    GtkAccelGroup *accel_group;

    toplevel = gtk_widget_get_toplevel(widget);
    if (gtk_widget_is_toplevel(toplevel)) {
        accel_group = gtk_ui_manager_get_accel_group(viewer->ui_manager);
        gtk_window_add_accel_group(GTK_WINDOW(toplevel), accel_group);
    }
}

static void
evbp_viewer_class_init(EvbpViewerClass *class)
{
    GObjectClass *object_class = G_OBJECT_CLASS(class);
    GtkWidgetClass *widget_class = GTK_WIDGET_CLASS(class);

    object_class->dispose = evbp_viewer_dispose;
    widget_class->hierarchy_changed = evbp_viewer_hierarchy_changed;
}

static const GtkActionEntry action_entries[] = {
    { "GoPreviousPage", GTK_STOCK_GO_UP,
      "_Previous Page", "<control>Page_Up",
      "Go to the previous page",
      G_CALLBACK(evbp_viewer_previous_page) },
    { "GoNextPage", GTK_STOCK_GO_DOWN,
      "_Next Page", "<control>Page_Down",
      "Go to the next page",
      G_CALLBACK(evbp_viewer_next_page) }
};

static const GtkToggleActionEntry toggle_action_entries[] = {
    { "ViewBestFit", EV_STOCK_ZOOM_PAGE,
      "_Best Fit", NULL,
      "Make the document fill the window",
      G_CALLBACK(evbp_viewer_zoom_best_fit), FALSE },
    { "ViewPageWidth", EV_STOCK_ZOOM_WIDTH,
      "Fit Page _Width", NULL,
      "Make the document fill the window width",
      G_CALLBACK(evbp_viewer_zoom_page_width), FALSE },
    { "ViewContinuous", EV_STOCK_VIEW_CONTINUOUS,
      "_Continuous", NULL,
      "Show the entire document",
      G_CALLBACK(evbp_viewer_toggle_continuous), FALSE },
};

static const GtkActionEntry accel_entries[] = {
    { "FocusPageSelector", NULL, "", "<control>l", NULL,
      G_CALLBACK(focus_page_selector) },
};

static const gchar ui_string[] =
    "<ui>\n"
    "  <toolbar name=\"ViewerToolbar\">\n"
    "    <toolitem name=\"GoPreviousPage\" action=\"GoPreviousPage\"/>\n"
    "    <toolitem name=\"GoNextPage\" action=\"GoNextPage\"/>\n"
    "    <separator/>\n"
    "    <toolitem name=\"PageSelector\" action=\"PageSelector\"/>\n"
    "    <separator/>\n"
    "    <toolitem name=\"ViewBestFit\" action=\"ViewBestFit\"/>\n"
    "    <toolitem name=\"ViewPageWidth\" action=\"ViewPageWidth\"/>\n"
    "    <toolitem name=\"ViewContinuous\" action=\"ViewContinuous\"/>\n"
    "  </toolbar>\n"
    "  <accelerator name=\"FocusPageSelector\" action=\"FocusPageSelector\"/>\n"
    "</ui>\n";

static void
evbp_viewer_init(EvbpViewer *viewer)
{
    GError *error = NULL;
    GtkWidget *toolbar;
    GtkAction *action;
    GList *actions;

    gtk_orientable_set_orientation(GTK_ORIENTABLE(viewer),
                                   GTK_ORIENTATION_VERTICAL);

    viewer->model = ev_document_model_new();

    viewer->action_group = gtk_action_group_new("ViewerActions");
    gtk_action_group_add_actions(viewer->action_group, action_entries,
                                 G_N_ELEMENTS(action_entries), viewer);
    gtk_action_group_add_toggle_actions(viewer->action_group,
                                        toggle_action_entries,
                                        G_N_ELEMENTS(toggle_action_entries),
                                        viewer);
    action = g_object_new(EV_TYPE_PAGE_ACTION,
                          "name", "PageSelector",
                          "label", "Page",
                          "tooltip", "Select Page",
                          "icon-name", "text-x-generic",
                          "visible-overflown", FALSE,
                          NULL);
    ev_page_action_set_model(EV_PAGE_ACTION(action), viewer->model);
    g_signal_connect(action, "activate-link",
                     G_CALLBACK(page_action_link_activated), viewer);
    gtk_action_group_add_action(viewer->action_group, action);
    g_object_unref(action);
    gtk_action_group_set_sensitive(viewer->action_group, FALSE);

    viewer->accel_group = gtk_action_group_new("ViewerAccelerators");
    gtk_action_group_add_actions(viewer->accel_group, accel_entries,
                                 G_N_ELEMENTS(accel_entries), viewer);
    gtk_action_group_set_sensitive(viewer->accel_group, FALSE);

    viewer->ui_manager = gtk_ui_manager_new();
    gtk_ui_manager_insert_action_group(viewer->ui_manager,
                                       viewer->action_group, 0);
    gtk_ui_manager_insert_action_group(viewer->ui_manager,
                                       viewer->accel_group, 1);
    if (!gtk_ui_manager_add_ui_from_string(viewer->ui_manager, ui_string,
                                           sizeof(ui_string) - 1, &error)) {
        g_warning("Failed to load ui from string: %s", error->message);
        g_error_free(error);
        error = NULL;
    }

    toolbar = gtk_ui_manager_get_widget(viewer->ui_manager, "/ViewerToolbar");
    gtk_box_pack_start(GTK_BOX(viewer), toolbar, FALSE, FALSE, 0);
    gtk_widget_show(toolbar);

    /* connect the accelerators since GtkUIManager only does that for menus */
    gtk_ui_manager_ensure_update(viewer->ui_manager);
    actions = gtk_action_group_list_actions(viewer->action_group);
    g_list_foreach(actions, (GFunc)gtk_action_connect_accelerator, NULL);
    g_list_free(actions);

    /* wire up toolbar callbacks */
    sizing_mode_handler(G_OBJECT(viewer->model), NULL, viewer);
    continuous_handler(G_OBJECT(viewer->model), NULL, viewer);
    g_signal_connect(viewer->model, "notify::sizing-mode",
                     G_CALLBACK(sizing_mode_handler), viewer);
    g_signal_connect(viewer->model, "notify::continuous",
                     G_CALLBACK(continuous_handler), viewer);

    viewer->scroll = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(viewer->scroll),
                                   GTK_POLICY_AUTOMATIC,
                                   GTK_POLICY_AUTOMATIC);
    gtk_widget_show(viewer->scroll);

    viewer->view = ev_view_new();
    g_signal_connect(viewer->view, "focus-in-event",
                     G_CALLBACK(view_focus_changed), viewer);
    g_signal_connect(viewer->view, "focus-out-event",
                     G_CALLBACK(view_focus_changed), viewer);
    ev_view_set_model(EV_VIEW(viewer->view), viewer->model);
    ev_view_set_loading(EV_VIEW(viewer->view), TRUE);
    gtk_widget_show(viewer->view);

    gtk_container_add(GTK_CONTAINER(viewer->scroll), viewer->view);
    gtk_box_pack_end(GTK_BOX(viewer), viewer->scroll, TRUE, TRUE, 0);
}

GtkWidget *
evbp_viewer_new(void)
{
    return g_object_new(EVBP_TYPE_VIEWER, NULL);
}

gboolean
evbp_viewer_load_uri(EvbpViewer *viewer, const gchar *uri, GError **error)
{
    if (viewer->document)
        g_object_unref(viewer->document);

    viewer->document = ev_document_factory_get_document(uri, error);
    g_return_val_if_fail(viewer->document, FALSE);

    g_debug("loading document \"%s\"", uri);
    ev_document_model_set_document(viewer->model, viewer->document);
    ev_view_set_loading(EV_VIEW(viewer->view), FALSE);
    gtk_action_group_set_sensitive(viewer->action_group, TRUE);

    return TRUE;
}
